-- Command menu. 
-- Displays command in a similar fashion as the main Turbine menu.
-- Commands are displayed bottom/up if the menu is opened in the bottom half of the display

-- HACK: If displaying bottom/up, the menu will shift upward as if SetTop() is actually SetBottom(). 
-- TODO: Whether the menu is displaying bottom/up or top/down should be externalized

import "Turbine.UI";
import "Turbine.UI.Lotro";
import "Equendil.UI.DecoratedContainer";
import "Equendil.Utils";
import "Equendil.PluginMng.KnownPlugins";

-- resources
local GradiantBlue = 0x41101B27;
local GradiantBlueReversed = 0x41101D9B;

local RingBackground = "Equendil/UI/images/ring_grey5.tga"
local RingIcon = 0x41005F1D;

--[[
GetPluginFavIcon() looks for an icon definition associated to a plugin
Specifically, it looks for a Favicon.lua file in the same folder as the 'package' of a plugin
if that exists, it expects to find a "Favicon" table with a 'Regular' subtable with the following fields:
- Resource: the relative path from Plugins/ of a targa or JPEG image, or the id of an in game resource
- Width: the width of the image (unused as of now)
- Height: the height of the image (unused as of now)

eg: in Favicon.lua
Favicon = { 
	Regular = { Resource = 0x41103547, Width = 25, Height = 25 },
};
--]]
local function GetPluginFavIcon( plugin )
	-- extract the 'path' of the 'package' of the plugin (which is setup in the .plugin file)
	local package = plugin:GetPackage();
	i, j = string.find( package, "[%w_%.]+%." ); -- extract "a.b.c." from "a.b.c.d"
	if i == nil then return nil; end
	local path = string.sub( package, i, j );
	
	-- import a "Favicon.lua" in the same place as the 'package' if it exists
	local success = pcall( import, path .. "Favicon" ); -- protected call to catch errors ...
	if not success then	
		return nil;	-- import will fail if the file does not exist or is malformed, time to return ...
	end

	-- Favicon.lua should now be imported, go down the namespaces where it should have been loaded
	local env = _G -- begin from the global environment
	for name in string.gfind( path, "[%w_]+" ) do -- for each 'name' in the path
		env = env[name]; -- go down one level
		if env == nil then -- this shouldn't occur, but just in case something went terribly wrong ...
			return nil;
		end
	end
	
	-- and extracts the information we need
	local icon = env.Favicon;
	if icon == nil or icon.Regular == nil then 
		return nil;
	end
	
	return icon.Regular.Resource;
end

-- Get an icon for a plugin. Looks first for a Favicon, else a defined icon from 'known' plugins, else returns a default icon
local function GetPluginIcon( plugin )
	local icon = GetPluginFavIcon( plugin );
	if icon == nil then
		icon = KnownPlugins.Icons[plugin:GetName()];
	end
	if icon == nil then
		icon = RingIcon;
	end
	
	return icon;
end

-- Individual controls displayed in the menu
local MenuItem = class( Turbine.UI.Control );
function MenuItem:Constructor( command )
	Turbine.UI.Control.Constructor( self );
	
	self.command = command;
	self.Clicked = Equendil.Utils.WeakEvent();
	self.mouseInside = false;
	self.executable = false;
	
	self:SetSize( 240, 33 );
	self:SetBackColor( Turbine.UI.Color( 1, 0, 0, 0 ) );
	-- self:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	-- self:SetBackground( GradiantBlue );

	-- adding a 'shortcut' quickslot so we can 'execute' commands not available through the shell
	self.quickslot = Turbine.UI.Lotro.Quickslot();
	self.quickslot:SetParent( self );
	self.quickslot:SetStretchMode( 1 );
	self.quickslot:SetAllowDrop( false );
	local shortcut = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, command:GetShellCommand() );
	self.quickslot:SetShortcut( shortcut );
	self.quickslot:SetBackColor( Turbine.UI.Color( 1, 1, 0, 0 ) );
	self.quickslot:SetSize( 240, 33 ); -- seems we have to set the size after SetShortcut() is called, or it resets somehow
	
	-- 'background' control covering the quickslot to hide it and catch mouse events in cases where we don't want the alias shortcut to get them
	self.background = Turbine.UI.Control();
	self.background:SetParent( self );
	self.background:SetSize( 240, 33 );
	self.background:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	self.background:SetBackground( GradiantBlue );
	
	-- setup a big label for the command name
	self.labelBackground = Turbine.UI.Control();
	self.labelBackground:SetParent( self );
	self.labelBackground:SetSize( 240, 32 );  -- we use a height one pixel less to leave a separation line between items from the gradiant.
	self.labelBackground:SetBackColorBlendMode( Turbine.UI.BlendMode.Overlay );
	self.labelBackground:SetBackColor( Equendil.UI.Color.Black ); -- cover the gradient save for 1 pixel 
	self.labelBackground:SetMouseVisible( false );
	
	self.label = Turbine.UI.Label();
	self.label:SetParent( self.labelBackground );
	self.label:SetSize(200,32);
	self.label:SetPosition( 40, 0 );
	self.label:SetFont( Turbine.UI.Lotro.Font.TrajanPro16 );
	self.label:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	self.label:SetFontStyle( Turbine.UI.FontStyle.Outline );
	self.label:SetForeColor( Equendil.UI.Color.White );
	self.label:SetMouseVisible( false );
	self.label:SetText( command:GetName() );

	-- setup an icon
	self.icon = Turbine.UI.Control();
	self.icon:SetParent( self );
	self.icon:SetMouseVisible( false );
	self.icon:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	local plugin = self.command:GetPlugin();
	if plugin ~= nil then
		local icon = GetPluginIcon( plugin );
		if icon ~= nil then
			-- we have an icon for this plugin, set its background, position and stretch it to fit 28x28 ( height of a menu entry - 2 px spacing on each side)
			local success = pcall( Turbine.UI.Control.SetBackground, self.icon, icon ); -- set the background through a protected call in case we have errors
			if success then
				self.icon:SetStretchMode( 2 ); -- will shrink/grow the size of the control to the size of the background image
				local width, height = self.icon:GetSize(); -- get the size of the image
				local scale = 26 / math.max( width, height ); -- find by how much we have to scale it
				width, height = width * scale, height * scale; -- compute the new width & height
				self.icon:SetPosition( ( 26 - width ) / 2 + 3, (26 - height ) / 2 + 3); -- center the icon
				self.icon:SetStretchMode( 1 ); -- normal stretch mode
				self.icon:SetSize( width, height ); -- and set its size
			end
		end
		-- 'ring' overlay that may mask parts of square icons to make them look round
		self.iconOverlay = Turbine.UI.Control();
		self.iconOverlay:SetParent( self );
		self.iconOverlay:SetBackground( RingBackground );
		self.iconOverlay:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
		self.iconOverlay:SetSize( 32, 32 );
		self.iconOverlay:SetMouseVisible( false );
		self.iconOverlay:SetStretchMode( 1 ); -- have to set the stretch mode like the icon or the blend mode won't work
	end

	-- actions on events
	local actionMouseEnter = function( sender, args )
		self.mouseInside = true; -- we want to remember where we are for MouseUp events
		self.labelBackground:SetBackColor( Equendil.UI.Color.Transparent ); -- Label is made transparent to uncover the gradient behind
	end
	
	local actionMouseLeave = function( sender, args ) 
		self.labelBackground:SetBackColor( Equendil.UI.Color.Black ); -- label made solid to hide the gradient behind (save for one pixel for separation)
		self.mouseInside = false;
	end
	
	local actionMouseDown = function( sender, args )
		self:SetBackground( GradiantBlueReversed );
	end
	
	local actionMouseUp = function( sender, args )
		self:SetBackground( GradiantBlue );
		if self.mouseInside then
			self.Clicked( args.Button ); -- fire a click event if the mouse is still inside the control
		end
	end

	-- we register actions on both the quickslot and the 'background' control, whichever has mouse visibility set to true will catch them
	self.quickslot.MouseEnter = actionMouseEnter;
	self.quickslot.MouseLeave = actionMouseLeave;
	self.quickslot.MouseDown = actionMouseDown;
	self.quickslot.MouseUp = actionMouseUp;
	
	self.background.MouseEnter = actionMouseEnter;
	self.background.MouseLeave = actionMouseLeave;
	self.background.MouseDown = actionMouseDown;
	self.background.MouseUp = actionMouseUp;
	
	-- setup a callback to update when the command changes
	self.updateAction = function() self:_Update(); end
	self.command.Changed:Add(self.updateAction); 
end

function MenuItem:SetupCommand()
	self.executable = self.command:IsFullyExecutable();
	-- if the command is executable we'll catch mouse events otherwise we'll leave them to the quickslot
	if self.executable then
		self.quickslot:SetMouseVisible( false );
		self.background:SetMouseVisible( true );
	else
		self.background:SetMouseVisible( false );
		self.quickslot:SetMouseVisible( true );
	end
	-- self.quickslot:SetSize( 240, 33 );
end

-- Garbage collector friendly 'dispose'
function MenuItem:Dispose()
	self:SetParent( nil );
	self.Clicked = nil;
	self.command.Changed:Remove( self.updateAction );
end

-- update the menu item 
function MenuItem:_Update()
	self.label:SetText( self.command:GetName() );
	
	local shortcut = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self.command:GetShellCommand() );
	self.quickslot:SetShortcut( shortcut );
	self.quickslot:SetSize( 240, 33 );
end

-- run the command therein
function MenuItem:Execute()
	-- if the command is not executable then the quickslot should have taken care of it
	self.command:Execute();
end

-- returns whether a menu item should be displayed
function MenuItem:IsHidden()
	return self.command:GetMode() == Equendil.PluginMng.Command.Mode.RunOnly and not self.command:GetPlugin():IsLoaded();
end

-- Main class
Menu = class( Turbine.UI.Window );
function Menu:Constructor()
	Turbine.UI.Window.Constructor( self );
	
	self.commands = {};
	self.menuItems = {};
	
	self:SetVisible( false );
	self:SetSize( 250, 33 );
	-- self:SetBackColor( Equendil.UI.Color.Black );
	
	self.borders = Equendil.UI.DecoratedContainer( Equendil.UI.Decoration.RoundSilver );
	self.borders:SetParent( self );
	self.borders:SetZOrder( 100 ); -- arbitrarily large value to keep the borders on top of the menu items so there's no nasty clipping
	self.borders:SetSize( self:GetSize() );
end

-- set the list of commands to manage
function Menu:Set( commands )
	self:Show( false );
	
	-- get rid of the old menu items
	for k, menuItem in pairs( self.menuItems ) do
		menuItem:Dispose();
	end
	self.menuItems = {};

	-- and build new ones
	self.commands = commands;
	for i, command in ipairs( commands ) do
		local menuItem = MenuItem( command );
		menuItem:SetParent( self );
		menuItem:SetLeft( 5 );
		menuItem:SetVisible( false );
		
		-- on click, execute the command and hide the menu. 
		-- We should probably fire an event instead for someone else to deal with the command, but let's keep this simple
		menuItem.Clicked = function( button )
			menuItem:Execute();
			self:Show(false);
		end
		table.insert( self.menuItems, menuItem );
	end

end	

-- Show/Hide the menu
function Menu:Show( flag )
	if not flag then 
		-- hide the menu
		self:Close();
		return;
	end
	
	-- find how many items need to be displayed
	local numVisible = 0;
	for i, menuItem in ipairs( self.menuItems ) do
		if not menuItem:IsHidden() then
			numVisible = numVisible + 1;
		end
	end

	-- set the height accordingly
	local height = 10 + 33 * numVisible;
	self:SetHeight( height );
	self.borders:SetHeight( height );
	
	-- initialize position and offset depending if we draw the menu top down or bottom up
	-- which is determined by the quadrant on the display in which we're displaying
	-- position of the menu is adjusted if displaying bottom up
	local topItem = 5;
	local offset = 33;
	if self:GetTop() > Turbine.UI.Display.GetHeight() / 2 then 
		-- menu is below the top half of the display, we'll display items bottom up
		offset = -33;
		topItem = height - 5 - 33;
		self:SetTop( self:GetTop() - height );
	end
	
	-- position items in the menu
	for i, menuItem in ipairs( self.menuItems ) do
		if not menuItem:IsHidden() then
			menuItem:SetTop( topItem );
			menuItem:SetupCommand();
			menuItem:SetVisible( true );
			topItem = topItem + offset;
		else
			menuItem:SetVisible( false );
		end
	end

	-- open the menu proper
	self:SetVisible( true );
	self:Activate();
end
